#undef DEBUG_DMA

#include <stdio.h>
#include <math.h>
#include <signal.h>
#include "dmavif1.h"
#include "screen.h" 
#include "geommath.h"
#include "camera.h"
#include "p2lassert.h"


// This toggles the VU1 and the CPU (via path3) version
#define FAST_VU_VERSION


#define FRONT_PLANE 10
#define BACK_PLANE  10000

int loop_again=1;

// this is executed when ctrl-c is pressed, ideally it stops the loop.
// the deinit functions have to be at the end of the main after the loop.
void sig_handle(int sig)
{
	loop_again=0;
}


// distribute sps2 specific global pointers to relevant classes
void  DistribGlobPtrs(void * pDescCached, void * pDescUncached)
{
	CDmaWrap::m_pDescCached = (sps2Memory_t *) pDescCached;
	CDmaWrap::m_pDescUncached = (sps2Memory_t *) pDescUncached;
}

// labels for vu1 particle-code
extern "C" void VU_vu1part_start(void);
extern "C" void VU_vu1part_end(void);

int iSPS2Device;
CScreen *cScreen=NULL;

void delete_screen()
{
	if (cScreen!=NULL) delete cScreen;
	cScreen=NULL;
}

// particle stuff
Mat44 mToScreen, mCamToScr;
void particles_init(void * dmaptr, uint32 uQSize);
void particles_update();
void UploadATexture(void * pDmaMem);

Mat44 fpuViewScreenMatrix(float scrz, float ax, float ay, 
	       float cx, float cy, float zmin, float zmax, float nearz, float farz)
{
	Mat44 m;
	float	az, cz;

	cz = (-zmax * nearz + zmin * farz) / (-nearz + farz);
	az  = farz * nearz * (-zmin + zmax) / (-nearz + farz);

	SetColumn(&m, 0, Vec4(ax*scrz, 0, 0, 0));
	SetColumn(&m, 1, Vec4(0, ay*scrz, 0, 0));
	SetColumn(&m, 2, Vec4(cx, cy, cz, 1));
	SetColumn(&m, 3, Vec4(0, 0, az, 0));

	return m;
}

int main()
{
	// Open sps2-library and setup PS2Linux stuff
	iSPS2Device = sps2Init();

	if (iSPS2Device<0) {
	  printf("Can't open SPS2 device.  Is the module loaded?\n");
	  exit(-1);
	}

	uint32 uAllocDmaMegs = 2;
	sps2Memory_t * pDescCached = sps2Allocate(uAllocDmaMegs*1024*1024, SPS2_MAP_BLOCK_4K | SPS2_MAP_CACHED, iSPS2Device);
	sps2Memory_t * pDescUncached = sps2Remap(pDescCached, SPS2_MAP_UNCACHED, iSPS2Device);
	DistribGlobPtrs(pDescCached, pDescUncached);
	void * pAllDmaMem = pDescCached->pvStart;
	int iVPU0; iVPU0=open("/dev/ps2vpu0", O_RDWR);
	int iVPU1; iVPU1=open("/dev/ps2vpu1", O_RDWR);
	for(int sig=0; sig<128; sig++)	// catch as many signals as possible
		signal(sig, sig_handle);

	
	// Setup Screen
	cScreen=new CScreen;
	atexit(delete_screen);
	cScreen->ScreenSetup();

	// Upload a texture
	UploadATexture(pAllDmaMem);

	// Setup a Camera
	float fScrWdth = cScreen->GetWidth();
	float fScrHgt = cScreen->GetHeight();
	float fDepthZ = cScreen->GetDepth();
	float fFov = 0.524f;	// total view of 60 degrees
	/*CCamera cCam;
	cCam.SetBitmapSize(fScrWdth, -fScrHgt, fDepthZ);
	cCam.SetViewPortOffset(2048, 2048, fDepthZ);
	cCam.SetClipPlanes(-10, -10000);
	cCam.SetEyePos(Vec3(0,0,1));
	float fMinU = -tanf(fFov);
	float fMinV = (fMinU*3)/4;
	cCam.SetBoundVRC(fMinU, fMinV, -fMinU, -fMinV);
	cCam.SetPosViewPlane(Vec3(0,0,600));
	cCam.SetRotViewPlane(Vec3(0,0,1), Vec3(0,1,0));
	mToScreen = cCam.GetWorldToBitmapTransform();
	SetRow(&mToScreen, 0, GetRow(mToScreen, 0)*16);
	SetRow(&mToScreen, 1, GetRow(mToScreen, 1)*16);*/

	// extra cam
	float fInvTan = 1.0f / tanf(fFov);
	float ax = (0.5f*fScrWdth) * fInvTan;
	float ay = ((-0.5f*fScrHgt*4) * fInvTan)/3;
	mCamToScr = fpuViewScreenMatrix(1.0f, ax, ay, 2048, 2048, 1, fDepthZ, FRONT_PLANE, BACK_PLANE);
	LoadIdentity(&mToScreen);
	SetColumn(&mToScreen, 3, Vec4(0,0,600, 1));
	mToScreen = mCamToScr * mToScreen;

	// copy to vu1micromem
	uint32 iLengthCodeInOcts = (((uint32)&VU_vu1part_end - (uint32)&VU_vu1part_start) + 7) >> 3;
	memcpy(VU1_MICRO_MEM, &VU_vu1part_start, iLengthCodeInOcts*8);
	
	const uint32 uQuads = (uAllocDmaMegs*1024*1024)/16;
	particles_init(pAllDmaMem, uQuads);	// offset pointer by sizeof screen

	// Flush Cache
	sps2FlushCache(iSPS2Device);

	while(loop_again)
	{
		// draw
		particles_update();

		// flip
		cScreen->Swap();
	}

	return 0;
}

// particle updating
#define PI 3.1415926535897932384626433832795f	
#define TWOPI (2*PI)

// RBUF_LEN is the amount of active particles permitted in a frame
#define RBUF_LEN	1200	// using CPU version this should be 1200 (since it's too slow)

// This is the amount of attempted births per frame
#define BIRTHS_PER_FRAME	8	// using CPU version this should be 8 (since it's too slow)

#ifdef FAST_VU_VERSION	// 25 times as many if this is the VU version
#undef RBUF_LEN
#undef BIRTHS_PER_FRAME
#define RBUF_LEN	30000
#define BIRTHS_PER_FRAME	200
#endif

struct SPart
{
	Vec3 vC1;
	float fStartTime;
	Vec3 vC2;
	float fInvTime;		// ((8-1)/(part.fEndTime-part.fStartTime))
};

struct SSpriteArrayParticle
{
	Vec3 vScaledGravity;

	float fAngleSpeed;
	float fAngleSpeedVel;
	float fAngleSpeedAcc;

	float fScale;
	float fScaleVel;
	float fScaleAcc;

	float fFriction;
};

SSpriteArrayParticle myHeader;

SPart * rbuf = NULL;
int32 iStartRBuf, iEndRBuf;
Vec4 Table[8+1];	// colorTable

float current;

void * chainbuf = NULL;
uint32 chainQSize;


void particles_init(void * dmaptr, uint32 uQSize)
{
	// alloc mem for ring buffer
	assert(RBUF_LEN*sizeof(SPart) < uQSize*16);
	rbuf = (SPart *) dmaptr;
	dmaptr = (void *) &((SPart *) dmaptr)[RBUF_LEN];
	uQSize -= ((RBUF_LEN*sizeof(SPart)+0xf)>>4);	
	
	
	
	assert(RBUF_LEN>2);
	chainbuf = dmaptr;
	chainQSize = uQSize;

	iStartRBuf = 0;
	iEndRBuf = RBUF_LEN-1;
	current = 0;

	// set all particles to be too old
	for(int i=0; i<RBUF_LEN; i++)
	{ rbuf[i].fStartTime=-1; rbuf[i].fInvTime=8; }

	myHeader.vScaledGravity = Vec3(0,-0.8f,0);
	myHeader.fAngleSpeed = 0.02f;
	myHeader.fAngleSpeedVel = 0.000003f;
	myHeader.fAngleSpeedAcc = 0.000001f;
	myHeader.fScale = 5;
	myHeader.fScaleVel = 0.01f;
	myHeader.fScaleAcc = 0.0f;
	myHeader.fFriction = 0.01f;
	
	// set Colors
	Table[0] = Vec4(127,0,0,0);
	Table[1] = Vec4(127,127,0,0);
	Table[2] = Vec4(0,127,0,0);
	Table[3] = Vec4(127,0,127,0);
	Table[4] = Vec4(0,127, 127,0);
	Table[5] = Vec4(0,63, 127,0);
	Table[6] = Vec4(0,127, 63,0);
	Table[7] = Vec4(0,33, 100,0);
	//Table[7] = Vec4(0,63,0,0);


	/*Table[0] = Vec4(53,30,25,15);
	Table[1] = Vec4(83,50,13,15);
	Table[2] = Vec4(60,27,00,33);
	Table[3] = Vec4(67,23,0,60);
	Table[4] = Vec4(55,17, 0,55);
	Table[5] = Vec4(60,7, 0,45);
	Table[6] = Vec4(54,5, 3,40);
	Table[7] = Vec4(45,35,32,38);*/
	
	Table[8] = Table[7];
}

void particles_life_death();
void particles_draw_cpu();
void particles_draw_VU();

float fRnd() { return ((float) (((int32) rand()) % ((int32) (RAND_MAX+1)))) / ((int32) RAND_MAX); }

void particles_update()
{
	// red is spawn/kill time used on the CPU
	// green is chain creation time on the CPU
	// blue is the Dma/VU/GS time
	DPUT_GS_BGCOLOR(0x0000ff);
	particles_life_death();
	DPUT_GS_BGCOLOR(0x00ff00);
#ifndef FAST_VU_VERSION
	// cpu version
	particles_draw_cpu();
	DPUT_GS_BGCOLOR(0xff0000);	// Created particles now render them
	while((*EE_D2_CHCR)&0x0100)
                ;
#else
	// VU version
	particles_draw_VU();
	DPUT_GS_BGCOLOR(0xff0000);	// Wait for chain launch to complete
	while((*EE_D1_CHCR)&0x0100)
                ;
#endif

	// rendering complete
	DPUT_GS_BGCOLOR(0x000000);
	
	current+=1.0f;
}

// kill old ones and spawn new ones
// in a circular buffer
void particles_life_death()
{
	static float vx = 0, vy = 0, vz = 0;
	vx -= 0.001; vy += 0.003; vz += 0.002;
	Mat33 m33; LoadRotation(&m33, vx, vy, vz);
	
	
	// kill the old ones
	bool bProceed;
	do
	{
		int32 iNxtEndRBuf = (iEndRBuf+1) % RBUF_LEN;
		bProceed = false;

		if(iNxtEndRBuf!=iStartRBuf && (current - rbuf[iEndRBuf].fStartTime)*rbuf[iEndRBuf].fInvTime > 7)
		{ bProceed = true; iEndRBuf=iNxtEndRBuf; } 
	}
	while(bProceed);

	// give birth
	static uint32 uDir = 0;
	int i=0;
	int iNxtSt;
	while(i<BIRTHS_PER_FRAME && 
		  (iNxtSt=(iStartRBuf+1)%RBUF_LEN) != iEndRBuf 
		  )
	{
		rbuf[iNxtSt].fStartTime = current;
		rbuf[iNxtSt].fInvTime = (8-1) / (60*6.0f);	// 3 seconds

		// just a curve
		rbuf[iNxtSt].vC1 = m33 * ((fRnd()*100+350)*Normalize(Vec3(fRnd()-0.5f,0.6f,fRnd()-0.5f)));
		if(rbuf[iNxtSt].vC1.y < 0) rbuf[iNxtSt].vC1.y *= (-1);
		if(rbuf[iNxtSt].vC1.z > 0) rbuf[iNxtSt].vC1.z *= (-1);
		rbuf[iNxtSt].vC2 = -rbuf[iNxtSt].vC1 + 250*GetColumn(m33, 0);

		// set a direction for the rotation
		*((uint32 *) &rbuf[iNxtSt].fInvTime) &= 0xfffffffe;
		*((uint32 *) &rbuf[iNxtSt].fInvTime) |= uDir;	// 0 is right and 1 is left
		iStartRBuf = iNxtSt;
		++i;
		uDir = 1 - uDir;	// toggle
	}

	// kill the old ones again!
	// this will only have an effect if the buffer 
	// was almost empty (less then 2 live ones)
	// and has just been refilled
	do
	{
		int32 iNxtEndRBuf = (iEndRBuf+1) % RBUF_LEN;
		bProceed = false;

		if(iNxtEndRBuf!=iStartRBuf && (current - rbuf[iEndRBuf].fStartTime)*rbuf[iEndRBuf].fInvTime > 7)
		{ bProceed = true; iEndRBuf=iNxtEndRBuf; } 
	}
	while(bProceed);
}

// utility functions for stitching pages
#define SYNC(dmaBase)	\
{	\
	const uint32 uCur = (uint32) (dmaBase).GetCurLocation();	\
	if((uCur&0xff0) == 0xff0)	\
	{	\
		(dmaBase).AttachSrcNextTag((void *) (uCur+0x10));	\
		(dmaBase).EndPacket();	\
	}	\
}

// Basicly just add a reftag pointing to the data but if the data area crosses pages then
// add a reftag for each page in use. Only insert the lower vif code in the first reftag.
void UnloadByRef(CDmaVif1* pDmaVif1, uint32 uVUaddr, void * ptr, uint32 uWrites, uint32 uLowerTTECode)
{	
	uint32 uPtr = (uint32) ptr;
	
	// while we have more work to do
	while(uWrites > 0)
	{
		// find page border	
		uint32 next_page = (uPtr+0x1000)&(~0xfff);	// a pointer to next page in line
		assert(next_page >= uPtr);
		uint32 uQDistToEnd = next_page - uPtr;
		assert((uQDistToEnd&0xf) == 0);
		uQDistToEnd >>= 4;		// amount of quads from start of data to end of current page

		// attach
		SYNC(*pDmaVif1);
		// del is the amount of quad writes we can/will do
		const uint32 del = (uWrites < uQDistToEnd) ? uWrites : uQDistToEnd;
		const uint64 uTTE = UNPACK_TO_TOP(V4_32, del, uVUaddr)<<32 | uLowerTTECode; 
		pDmaVif1->AttachSrcRefTag((void *) uPtr, del, uTTE);
		uPtr += (del*16);
		uVUaddr += del;
		uWrites -= del;
		uLowerTTECode = NOP;
	}
}

// VU asm version of the code (dmachain creation)
void particles_draw_VU()
{
	const uint32 MAX_NUM_PARTS = 45;
	const uint32 POD_PART_SZE = 2;
	const uint64 GLB_PART_DATA_PTR = 1024 - 28;

	// Flush Cache
	sps2FlushCache(iSPS2Device);

	// page align dmachain so we don't have to worry about
	// stitching for the entrycode, this way stitching is only needed
	// for the actual reftags in the loop
	assert((((uint32) chainbuf)&0xf) == 0);
	void * dma_buf = (void *) ((((uint32) chainbuf)+0xfff)&(~0xfff));	// page align
	uint32 uBSize = ((uint32) dma_buf) - ((uint32) chainbuf);
	assert((uBSize&0xf) == 0);
	CDmaVif1 dmaVif1((void *) dma_buf, chainQSize - (uBSize>>4));
	dmaVif1.AttachSrcCntTag();
	
	// setup Vif registers
	dmaVif1.add64(OFFSET(GLB_PART_DATA_PTR>>1)<<32 | BASE(0));
	dmaVif1.add64(UNPACK_TO_TOP(V4_32, 25, 0)<<32 | STCYCL(1,1));

	// add matrix to chain
	const Vec4 c1 _ALIGN16 = GetColumn(mToScreen, 0);
	dmaVif1.add128(*((uint128 *) &c1));
	const Vec4 c2 _ALIGN16 = GetColumn(mToScreen, 1);
	dmaVif1.add128(*((uint128 *) &c2));
	const Vec4 c3 _ALIGN16 = GetColumn(mToScreen, 2);
	dmaVif1.add128(*((uint128 *) &c3));
	const Vec4 c4 _ALIGN16 = GetColumn(mToScreen, 3);
	dmaVif1.add128(*((uint128 *) &c4));

	// add grav+fric to chain
	SSpriteArrayParticle* pHeader=&myHeader;
	const Vec3 vScaledGravity=pHeader->vScaledGravity;
	const float fFriction=pHeader->fFriction;
	dmaVif1.add32(*((uint32 *) &vScaledGravity.x));
	dmaVif1.add32(*((uint32 *) &vScaledGravity.y));
	dmaVif1.add32(*((uint32 *) &vScaledGravity.z));
	dmaVif1.add32(*((uint32 *) &fFriction));

	// angle + cur
	const float fAngleSpeed=(1.0f/TWOPI)*pHeader->fAngleSpeed;	// 3e22f983
	const float fAngleSpeedVel=(1.0f/TWOPI)*pHeader->fAngleSpeedVel;
	const float fAngleSpeedAcc=(0.5f/TWOPI)*pHeader->fAngleSpeedAcc;
	const float fCurrentTime=current;
	dmaVif1.add32(*((uint32 *) &fAngleSpeed));
	dmaVif1.add32(*((uint32 *) &fAngleSpeedVel));
	dmaVif1.add32(*((uint32 *) &fAngleSpeedAcc));
	dmaVif1.add32(*((uint32 *) &fCurrentTime));

	// deliver scale
	const float fScaleStart=pHeader->fScale;
	const float fScaleVel=pHeader->fScaleVel;
	const float fScaleAcc=0.5f*pHeader->fScaleAcc;
	dmaVif1.add32(*((uint32 *) &fScaleStart));
	dmaVif1.add32(*((uint32 *) &fScaleVel));
	dmaVif1.add32(*((uint32 *) &fScaleAcc));
	dmaVif1.add32(0);
	
	// Deliver Clip values
	const float fLft = 2048 - (((int32) cScreen->GetWidth())/2);
	const float fRgt = 2048 + (((int32) cScreen->GetWidth())/2)-1;
	const float fLow = 2048 - (((int32) cScreen->GetHeight())/2);
	const float fHgh = 2048 + (((int32) cScreen->GetHeight())/2)-1;
	float clip[4] _ALIGN16 = {fLft, fLow, fRgt, fHgh};
	dmaVif1.add128(*((uint128 *) clip));

	// Deliver Front, Back, aspX, aspY. CamToScreen[0][0] and CamToScreen[1][1]
	const float front = FRONT_PLANE;
	const float back = BACK_PLANE;
	const float ax = mCamToScr.m_fMat[0*4+0];
	const float ay = mCamToScr.m_fMat[1*4+1];
	dmaVif1.add32(*((uint32 *) &front));
	dmaVif1.add32(*((uint32 *) &back));
	dmaVif1.add32(*((uint32 *) &ax));
	dmaVif1.add32(*((uint32 *) &ay));

	// Giftag of a strip
	dmaVif1.add64( GIFTAG(0, 0, 1, PRIM_SET(4, 0, 1, 0, 0, 0, 0, 0, 0), 0, 9) );
	dmaVif1.add64( GIFTAG_REGS12(ST, RGBAQ, XYZ2, ST, XYZ2, ST, XYZ2, ST, XYZ2, 0, 0, 0) );

	// add colors
	for(int i=0; i<8; i++)
		dmaVif1.add128(*((uint128 *) &Table[i]));

	// add texture registers to be kicked here
	dmaVif1.add64( GIFTAG(6, 1, 0, 0, 0, 1) );
	dmaVif1.add64( GIFTAG_REGS1(AD) );

	/*dmaVif1.add64( ALPHA_SET(0, 2, 0, 1, 127) );
	dmaVif1.add64( ALPHA_1 );

	dmaVif1.add64( TEST_SET(0, 0, 0, 0, 0, 0, 1, 1) );
	dmaVif1.add64( TEST_1 );*/

	
	//for(int q=2; q<6; q++)
	for(int q=0; q<6; q++)
	{
		dmaVif1.add64(0);
		dmaVif1.add64(0xf);
	}
	dmaVif1.EndPacket();

	// find out how many are active
	int32 iNumLiveParts = (iStartRBuf > iEndRBuf) ? (iStartRBuf-iEndRBuf)+1 : (RBUF_LEN + (iStartRBuf-iEndRBuf))+1;

	// not a nice trick but I do it as an 
	// easy way to avoid sending off empty buffers
	if(iNumLiveParts == 2) return;
	assert(iNumLiveParts > 2);

	// entries to jump table
	const uint32 part_init = 0;
	const uint32 part_draw = 2;

	// setup buffers of living particles
	SPart * buf1, * buf2;
	int32 len1, len2;
	if(iStartRBuf > iEndRBuf) 
	{
		// range: [iEndRBuf, iStartRBuf]
		buf1 = (SPart *) &(rbuf[iEndRBuf]);
		buf2 = (SPart *) NULL;
		len1 = iNumLiveParts;
		len2 = 0;
	}
	else
	{
		// range: [0; iStartRBuf] U [iEndRBuf; RBUF_LEN)
		buf1 = (SPart *) &(rbuf[iEndRBuf]);
		buf2 = (SPart *) &(rbuf[0]);
		len1 = RBUF_LEN-iEndRBuf;
		len2 = iStartRBuf+1;
	}

	// create chain
	uint32 uCurCall = MSCALL(part_init);
	for(int k=0; k<2; k++)
	{
		SPart * buf = (k==0) ? buf1 : buf2;
		int32 len = (k==0) ? len1 : len2;

		// unload all perfect chunks to chain
		while(len >= (int32) MAX_NUM_PARTS)
		{
			//const uint64 uTTE = UNPACK_TO_TOP(V4_32, 2*MAX_NUM_PARTS, POD_PART_SZE+0)<<32 | uCurCall;
			//dmaVif1.AttachSrcRefTag(buf, 2*MAX_NUM_PARTS, uTTE);
			UnloadByRef(&dmaVif1, POD_PART_SZE+0, (void *) buf, 2*MAX_NUM_PARTS, uCurCall);

			SYNC(dmaVif1);
			dmaVif1.AttachSrcCntTag(((uint64) MAX_NUM_PARTS)<<32 | UNPACK_TO_TOP_UNSIGNED(V4_8, 1, 0));
			dmaVif1.EndPacket();

			buf += MAX_NUM_PARTS;
			len -= MAX_NUM_PARTS;
			uCurCall = MSCALL(part_draw);
		}

		// if there's anything left in the buffer
		if(len > 0)
		{
			//const uint64 uTTE1 = UNPACK_TO_TOP(V4_32, 2*len, POD_PART_SZE+0)<<32 | uCurCall;
			//dmaVif1.AttachSrcRefTag(buf, 2*len, uTTE1);
			UnloadByRef(&dmaVif1, POD_PART_SZE+0, (void *) buf, 2*len, uCurCall);
			
			// if this was buffer1 (k==0)
			// then fuse end of current chunk with start of buffer2
			if(k==0 && len2>0)
			{
				// for a perfect buffer
				const int32 iMissingParts = MAX_NUM_PARTS - len;
				const int32 del2 = (len2 < iMissingParts) ? len2 : iMissingParts;

				//const uint64 uTTE2 = UNPACK_TO_TOP(V4_32, 2*del2, POD_PART_SZE+len*2)<<32 | NOP;
				//dmaVif1.AttachSrcRefTag(buf2, 2*del2, uTTE2);
				UnloadByRef(&dmaVif1, POD_PART_SZE+len*2, (void *) buf2, 2*del2, NOP);

				buf2 += del2;
				len += del2;
				len2 -= del2;
			}

			// add a call
			SYNC(dmaVif1);			
			dmaVif1.AttachSrcCntTag(((uint64) len)<<32 | UNPACK_TO_TOP_UNSIGNED(V4_8, 1, 0));
			dmaVif1.EndPacket();
			uCurCall = MSCALL(part_draw);
		}
	}
	SYNC(dmaVif1);
	dmaVif1.AttachSrcEndTag(uCurCall);
	dmaVif1.EndPacket();


	dmaVif1.Launch();
}

// more utility functions for stitching pages
uint32 uCurLoc;

void InsertNxtTag(CDmaWrap * pDmaBase, bool bFirstInsert=false)
{
	if(!bFirstInsert) pDmaBase->EndPacket();
	uCurLoc = (uint32) pDmaBase->GetCurLocation();
	uint32 uNxtAddr = (uCurLoc&(~0xfff))+0x1000;
	pDmaBase->AttachSrcNextTag((void *) uNxtAddr);
}

#define VERIFY_LOC(dmaBase) \
{	\
	uint32 uCurL = (uint32) dmaBase.GetCurLocation();	\
	if((uCurL&0xfff) == 0) InsertNxtTag(&dmaBase);	\
}

void Terminate(CDmaWrap * pDmaBase)
{
	pDmaBase->EndPacket();
	int32 bSize = (((uint32) pDmaBase->GetCurLocation()) - uCurLoc) - 0x10;
	assert(bSize >= 0);
	assert((bSize&0xf)==0);
	uint32 uQSize = (uint32) (bSize>>4);
	((uint64 *) uCurLoc)[0] = GEN_END_TAG(uQSize);
}

// render particles for C++ version over path3
#include "dmagif.h"

float Abs(float x) { return (x < 0) ? -x : x; }
Vec3 AbsV(Vec3 v) { return Vec3(Abs(v.x), Abs(v.y), Abs(v.z)); }
Vec3 prod(Vec3 v1, Vec3 v2) { return Vec3(v1.x*v2.x, v1.y*v2.y, v1.z*v2.z); }
Vec4 MAXxyw(const Vec4 &v1, const Vec4 &v2)
{
	const float x = (v1.x>v2.x)?v1.x:v2.x;
	const float y = (v1.y>v2.y)?v1.y:v2.y;
	const float w = (v1.w>v2.w)?v1.w:v2.w;
	return Vec4( x, y, 0, w);
}
Vec4 MINIxyw(const Vec4 &v1, const Vec4 &v2)
{
	const float x = (v1.x<v2.x)?v1.x:v2.x;
	const float y = (v1.y<v2.y)?v1.y:v2.y;
	const float w = (v1.w<v2.w)?v1.w:v2.w;
	return Vec4( x, y, 0, w);
}
int16 FMANDxyw(const Vec4 &v)
{
	// emulated fmand by 0xd0
	// makes a sign check on xyw
	int16 r;
	r = (v.x<0)?0x80:0;
	r |= ( (v.y<0)?0x40:0 );
	r |= ( (v.w<0)?0x10:0 );
	return r;
}
// simulated EFU instruction, expf() is too slow
float EEXP(float x)
{
	//return expf(-x);

	// Constants stolen from the VU manual
	float E1; *((uint32 *) &E1) = 0x3e7fffa8;
	float E2; *((uint32 *) &E2) = 0x3d0007f4;	
	float E3; *((uint32 *) &E3) = 0x3b29d3ff;
	float E4; *((uint32 *) &E4) = 0x3933e553;
	float E5; *((uint32 *) &E5) = 0x36b63510;
	float E6; *((uint32 *) &E6) = 0x353961ac;

	// Horners rule
	float tmp = 1.0f+(E1+(E2+(E3+(E4+(E5+E6*x)*x)*x)*x)*x)*x;
	tmp*=tmp; tmp*=tmp;

	return 1.0f / tmp;
}

// CPU version
void particles_draw_cpu()
{
	SSpriteArrayParticle* pHeader=&myHeader;

	const float fCurrentTime=current;
	const float fFriction=pHeader->fFriction;

	const float fAngleSpeed=(1.0f/TWOPI)*pHeader->fAngleSpeed;	// 3e22f983
	const float fAngleSpeedVel=(1.0f/TWOPI)*pHeader->fAngleSpeedVel;
	const float fAngleSpeedAcc=(1.0f/TWOPI)*0.5f*pHeader->fAngleSpeedAcc;

	const float ax = mCamToScr.m_fMat[0*4+0]*16;
	const float ay = (mCamToScr.m_fMat[1*4+1]*16) / ax;
	
	const float fScaleStart=ax*pHeader->fScale;
	const float fScaleVel=ax*pHeader->fScaleVel;
	const float fScaleAcc=ax*0.5f*pHeader->fScaleAcc;
	
	const Vec3 vScaledGravity=pHeader->vScaledGravity;

	// setup Taylor
	Vec3 s1; *((uint32 *) &s1.x)=0x40c90fdb; *((uint32 *) &s1.y)=0x40c90fdb; s1.z=0;
	Vec3 s2; *((uint32 *) &s2.x)=0xc2255de1; *((uint32 *) &s2.y)=0xc2255de1; s2.z=0;
	Vec3 s3; *((uint32 *) &s3.x)=0x42a33458; *((uint32 *) &s3.y)=0x42a33458; s3.z=0;
	Vec3 s4; *((uint32 *) &s4.x)=0xc2992663; *((uint32 *) &s4.y)=0xc2992663; s4.z=0;
	Vec3 s5; *((uint32 *) &s5.x)=0x421ed7bb; *((uint32 *) &s5.y)=0x421ed7bb; s5.z=0;

	// setup WorldToCam
	Vec4 c1 = GetColumn(mToScreen, 0);
	Vec4 c2 = GetColumn(mToScreen, 1);
	Vec4 c3 = GetColumn(mToScreen, 2);
	Vec4 c4 = GetColumn(mToScreen, 3);

	// a fix to make the cast float_to_int work like FTOI4
	// this is ofcourse specific to the cpu version
	c1.x *= 16; c1.y *= 16;
	c2.x *= 16; c2.y *= 16;
	c3.x *= 16; c3.y *= 16;
	c4.x *= 16; c4.y *= 16;

	// setup clip vectors
	const float fLft = 2048 - (((int32) cScreen->GetWidth())/2);
	const float fRgt = 2048 + (((int32) cScreen->GetWidth())/2)-1;
	const float fLow = 2048 - (((int32) cScreen->GetHeight())/2);
	const float fHgh = 2048 + (((int32) cScreen->GetHeight())/2)-1;
	
	// the .xy of the ClipVectors must also be scaled by 16
	const Vec4 vClipTop(fRgt*16, fHgh*16, 0, BACK_PLANE);
	const Vec4 vClipBot(fLft*16, fLow*16, 0, FRONT_PLANE);

	// calculate length
	int32 iNumLiveParts = (iStartRBuf > iEndRBuf) ? (iStartRBuf-iEndRBuf)+1 : (RBUF_LEN + (iStartRBuf-iEndRBuf))+1;

	// not a nice trick but I do it as an 
	// easy way to avoid sending off empty buffers
	if(iNumLiveParts == 2) return;
	assert(iNumLiveParts > 2);

	// setup the chain to path3
	assert((((uint32) chainbuf)&0xf) == 0);
	CDmaGif dmaGif((void *) chainbuf, chainQSize);
	InsertNxtTag(&dmaGif, true);

	uint32 nloop = iNumLiveParts;
	uint32 prim = PRIM_SET(4, 0, 1, 0, 0, 0, 0, 0, 0);
	VERIFY_LOC(dmaGif);
	dmaGif.add64(GIFTAG(nloop, 1, 1, prim, 0, 9));
	dmaGif.add64(GIFTAG_REGS12(ST, RGBAQ, XYZ2, ST, XYZ2, ST, XYZ2, ST, XYZ2, 0, 0, 0));

	const Vec4 st1(1,1,1,0) _ALIGN16;
	const Vec4 st2(0,1,1,0) _ALIGN16;
	const Vec4 st3(0,0,1,0) _ALIGN16;
	const Vec4 st4(1,0,1,0) _ALIGN16;

	// setup buffers of living particles
	SPart * buf1, * buf2;
	int32 len1, len2;
	if(iStartRBuf > iEndRBuf) 
	{
		// range: [iEndRBuf, iStartRBuf]
		buf1 = (SPart *) &(rbuf[iEndRBuf]);
		buf2 = (SPart *) NULL;
		len1 = iNumLiveParts;
		len2 = 0;
	}
	else
	{
		// range: [0; iStartRBuf] U [iEndRBuf; RBUF_LEN)
		buf1 = (SPart *) &(rbuf[iEndRBuf]);
		buf2 = (SPart *) &(rbuf[0]);
		len1 = RBUF_LEN-iEndRBuf;
		len2 = iStartRBuf+1;
	}

	for(int32 k=0; k<2; k++)
	{
		SPart * buf = (k==0) ? buf1 : buf2;
		int32 len = (k==0) ? len1 : len2;

		// I've tried to stay as true as possible to the VU version of the rendering 
		// code so this might seem like a highlevel description of the vu code.
		for(int i=0; i<len; i++)
		{
			SPart &part = buf[i];

			float fAge=fCurrentTime-part.fStartTime;	// particle age
			const int32 iDir = *((int32 *) &part.fInvTime);	// direction of rotation specified by the bottom bit of the float
			float fDir=(float) (1-((iDir&1)<<1));
		
			// calculate the following:
			// .x = age^2
			// .y = age*fDir
			// .z = age*friction
			// .w = age*7/lifespan, used to colorblend (from a collection of 8 colors)
			Vec4 vMuls, vM1(fAge, fAge, fAge, fAge), vM2(fAge, fDir, fFriction, part.fInvTime );
			vMuls = Vec4(vM1.x*vM2.x, vM1.y*vM2.y, vM1.z*vM2.z, vM1.w*vM2.w);//vM1 prod vM2;
			float fEEXP = EEXP(vMuls.z);	// EEXP(age*fFriction)
		
			// Calculate angle and scale
			Vec3 v;
			v.y    = fAngleSpeed+ fAngleSpeedVel*fAge+fAngleSpeedAcc*vMuls.x;
			float fScale = fScaleStart+fScaleVel*fAge+     fScaleAcc*vMuls.x;
			v.y *= vMuls.y;	// apply age*fDir to current anglespeed to get current angle

			// Get position from curve
			// p' = C1 + C2*e^(-age*friction) + grav*age
			Vec3 p = part.vC1 + part.vC2*fEEXP + vScaledGravity*fAge;

			// Calculate color
			float fOffs= vMuls.w;	// fAge*7/(endtime-starttime)
			int32 iOffs1 = (int32) fOffs;
			int32 iOffs2 = iOffs1+1;
			float t = fOffs - (float) iOffs1;

			//assert(iOffs1>=0 && iOffs1<8);
			//assert(iOffs2>0 && iOffs2<=8);
			//assert(t>=0 && t<1);
			Vec4 col = Table[iOffs1]+Table[iOffs2]*t-Table[iOffs1]*t;	// morph

			// convert angles into the range of -0.25; 0.25 (-PIHALF; PIHALF)
			// I'm only using the fields xy for this
			v.x=v.y-0.25f;	// y is the angle for the cosine and x is for the sine (but sine handled as a cosine)
			v = AbsV(v);
			Vec3 b; *((uint32 *) &b.x) = 0xcb000000; *((uint32 *) &b.y) = 0xcb000000; b.z = 0;
			Vec3 acc = b - v;	
			acc -= b*1.0f;	// acc = (b-v)-b becomes -IntOf( v ), meaning no decimals left
			acc += v*1.0f;	// by adding origional v to minus int_of( v ) we are left with just the decimals

			// cos(v)=-cos(v-0.5)=-cos(-(v-0.5)) = -cos(abs(v-0.5)) = 
			// -sin(abs(v-0.5)+0.25) = sin(abs(v-0.5)-0.25)
			// since v was in the range [0; 1) then
			// v' = abs(v-0.5)-0.25 will be in the range [-0.25; 0.25)
			// which is great for the precission when using the Taylor
			v = acc - Vec3(1.0f*0.5f, 1.0f*0.5f, 0)*1.0f;	// msub
			v = AbsV(v) - Vec3(0.25f, 0.25f, 0);

			// calculate the sine and cosine using Taylor for a sine
			Vec3 v2 = prod(v,v);
			Vec3 t1 = prod(v,v2);
			acc = prod(s1,v);		// mula
			Vec3 t2 = prod(t1,v2);
			acc += prod(s2,t1);	//	madda
			t1 = prod(t2,v2);
			acc += prod(s3,t2);	// madda
			t2 = prod(t1,v2);
			acc += prod(s4,t1);	// madda
			Vec3 r = acc + prod(s5,t2);	// madd
		

			// Create sprite direction vectors in pre-pixelspace (before the divide)
			Vec4 vDown = fScale * Vec4(r.x, r.y, 0, 0);
			Vec4 vHorz(vDown.y, ay*vDown.x, 0, 0);
			vDown.y *= (-ay);

			// transform point to pre-pixelspace and add half
			// the direction vectors to form first vertex.
			// then create the remaining 3 vertices
			Vec4 p1 = c4 + c1*p.x + c2*p.y + c3*p.z + vHorz*0.5f + vDown*0.5f;
			Vec4 p2 = p1 - vHorz;
			Vec4 p4 = p1 - vDown;
			Vec4 p3 = p2 - vDown;

			// project vertices but spare w for further clip testing
			const float q = 1.0f / p1.w;
			p1.x *= q; p2.x *= q; p3.x *= q; p4.x *= q;
			p1.y *= q; p2.y *= q; p3.y *= q; p4.y *= q;
			p1.z *= q; p2.z *= q; p3.z *= q; p4.z *= q;
			
			// clip test
			// if_xyw(ClipTop<min_vals || max_vals<ClipBot) sprite_invisible;
			const Vec4 vMax = MAXxyw( MAXxyw( MAXxyw(p1, p2), p3), p4);
			const Vec4 vMin = MINIxyw( MINIxyw( MINIxyw(p1, p2), p3), p4);
			const Vec4 vSigns1 = vClipTop - vMin;
			const Vec4 vSigns2 = vMax - vClipBot;
			const int16 fmand_res1 = FMANDxyw(vSigns1);	// fetch signs of field xyw
			const int16 fmand_res2 = FMANDxyw(vSigns2);

			// only renderable result is when, fmandres1|fmandres2 == 0
			// adc_bit (bit15) is enabled in all other possible cases
			const int16 adc_bit = (fmand_res1|fmand_res2)+(0x8000-0x10);

			// Add data to the chain
			VERIFY_LOC(dmaGif);
			dmaGif.add128(*((uint128 *) &st1));
			VERIFY_LOC(dmaGif);
			dmaGif.add32((int32) col.x); dmaGif.add32((int32) col.y); dmaGif.add32((int32) col.z); dmaGif.add32((int32) col.w);
			VERIFY_LOC(dmaGif);
			dmaGif.add32((int32) p1.x); dmaGif.add32((int32) p1.y); dmaGif.add32((int32) p1.z); dmaGif.add32(0x8000);

			VERIFY_LOC(dmaGif);
			dmaGif.add128(*((uint128 *) &st2));
			VERIFY_LOC(dmaGif);
			dmaGif.add32((int32) p2.x); dmaGif.add32((int32) p2.y); dmaGif.add32((int32) p2.z); dmaGif.add32(0x8000);
			VERIFY_LOC(dmaGif);
			dmaGif.add128(*((uint128 *) &st4));
			VERIFY_LOC(dmaGif);
			dmaGif.add32((int32) p4.x); dmaGif.add32((int32) p4.y); dmaGif.add32((int32) p4.z); dmaGif.add32(adc_bit);
			VERIFY_LOC(dmaGif);
			dmaGif.add128(*((uint128 *) &st3));
			VERIFY_LOC(dmaGif);
			dmaGif.add32((int32) p3.x); dmaGif.add32((int32) p3.y); dmaGif.add32((int32) p3.z); dmaGif.add32(adc_bit);
		}
	}

	Terminate(&dmaGif);
	
	dmaGif.Launch();
}


/************************************************************************************
************************ Texture Hack to prove it works aswell **********************
************************************************************************************/

void UploadATexture(void * pDmaMem)
{
	// pDmaMem is 16 quads from start of page
	// we're going to build the chain at pDmaMem
	// and then put a dummy texture of 64x16 in 32bit (4096bytes)
	// at the next page. This way we won't have to stitch
	uint32 next_page = (((uint32) pDmaMem)+0x1000)&(~0xfff);	
	for(int y=0; y<16; y++)	// Coder texture times T
		for(int x=0; x<64; x++)
		{
			uint32 col=0xff;

			if(y>=2 && y<5)
			{
				if(x>=10 && x<=(63-10))
				{
					float t = Abs(31.5f-x)/21.5f;
					t = (3-2*t)*t*t;
					col = ((int32) (t*0x5a+0x40));
				}
			}
			else if(y>=5 && y<=13)
			{
				if(x>=26 && x<=(63-26))
				{
					float t = Abs(31.5f-x)/5.5f;
					t = (3-2*t)*t*t;
					col = ((int32) (t*0x5a+0x40));
				}
			}
				
			((uint32 *) next_page)[y*64+x]=(col<<16)|(col<<8)|col;
		}
	
	// Flush Cache
	sps2FlushCache(iSPS2Device);

	// Create DmaChain
	CDmaVif1 dmaVif1(pDmaMem, 240);	// not 256 quads since 16quads were given to screen.cpp

	uint32 uDstPtr = (cScreen->GetFirstFreeGsPage()*2048)/64;//0x1a40;
	uint32 uQSize = (64*16*4)/16;

	dmaVif1.AttachSrcCntTag(DIRECT(6)<<32);
	dmaVif1.add64(GIFTAG(4, 0, 0, 0, 0, 1));	// GIFTAG(NLOOP, EOP, PRE, PRIM, FLG, NREG)
	dmaVif1.add64(GIFTAG_REGS1(AD));
	dmaVif1.add64(BITBLTBUF_SET(0, 0, 0, uDstPtr, 1, 0));		// 0x41a40 00000000L
	dmaVif1.add64(BITBLTBUF);
	dmaVif1.add64(TRXPOS_SET(0, 0, 0, 0, 0));
	dmaVif1.add64(TRXPOS);
	dmaVif1.add64(TRXREG_SET(64, 16));
	dmaVif1.add64(TRXREG);
	dmaVif1.add64(TRXDIR_SET(0));
	dmaVif1.add64(TRXDIR);
	dmaVif1.add64(GIFTAG(uQSize, 0, 0, 0, 2, 0));	// GIFTAG(NLOOP, EOP, PRE, PRIM, FLG, NREG)
	dmaVif1.add64(0);
	dmaVif1.EndPacket();

	//for(int page=0; page<((64*16*4)/4096); page++)
		//dmaVif1.AttachSrcRefTag(((char *) next_page)+page*4096, 4096/16, DIRECT((4096/16))<<32);
	dmaVif1.AttachSrcRefTag(((char *) next_page), 4096/16, DIRECT((4096/16))<<32);

	uint64 uNloop = 10;

	dmaVif1.AttachSrcEndTag(DIRECT(uNloop+1)<<32);
	dmaVif1.add64(GIFTAG(uNloop, 1, 0, 0, 0, 1));	// GIFTAG(NLOOP, EOP, PRE, PRIM, FLG, NREG)
	dmaVif1.add64(GIFTAG_REGS1(AD));
	dmaVif1.add64(TEST_SET(0, 0, 0, 0, 0, 0, 1, 2));	// 0x70000
	dmaVif1.add64(TEST_1);
	dmaVif1.add64(PABE_SET(0));
	dmaVif1.add64(PABE);
	dmaVif1.add64(TEXA_SET(0x7f, 1, 0x81));	//     0x810000807fL
	dmaVif1.add64(TEXA);
	dmaVif1.add64(FBA_SET(0));
	dmaVif1.add64(FBA_1);
	dmaVif1.add64(0);		// NOP data always 0
	dmaVif1.add64(0xf);	// NOP instruction
	dmaVif1.add64(TEX1_SET(0, 0, 3, 0, 0, 0, 0));		// 0x60
	dmaVif1.add64(TEX1_1);
	dmaVif1.add64(TEX0_SET(uDstPtr, 1, 0, 6, 4, 1, 0, 0, 0, 0, 0, 0));	// 0x620011a40L
	dmaVif1.add64(TEX0_1);
	dmaVif1.add64(CLAMP_SET(0,0,0,0,0,0));
	dmaVif1.add64(CLAMP_1);
	dmaVif1.add64(COLCLAMP_SET(0x1));
	dmaVif1.add64(COLCLAMP);
	dmaVif1.add64(0);
	dmaVif1.add64(TEXFLUSH);
	dmaVif1.EndPacket();
	dmaVif1.Launch();
	dmaVif1.Wait();
}
